unit n2_splines;

interface

uses n2_vector,n2_matrix;

type tsplineset=record
	   vy,wur:tvector;
	   end;
type tsplinesetx=record
	   vx,vy,wur:tvector;
	   end;

var splinewarning:byte;

function prepspline(vy:tvector):tsplineset;
function getsplineprep(splineset:tsplineset;x:double):double;
function getspline(vy:tvector;x:double):double;
function prepsplinex(vx,vy:tvector):tsplinesetx;
function getsplineprepx(splineset:tsplinesetx;x:double):double;
function getsplinex(vx,vy:tvector;x:double):double;
function getspline2(my:tmatrix;x,y:double):double;

implementation

uses n2_mathe;

function prepspline(vy:tvector):tsplineset;
var mat:tmatrix;
	eq,wur:tvector;
    i,j:longint;
	splineset:tsplineset;
begin
  mat:=initm(length(vy)-2,length(vy)-2);
  eq:=initv(length(vy)-2);
  wur:=initv(length(vy)-2);

  for j:=0 to length(vy)-3 do for i:=0 to length(vy)-3 do mat[j,i]:=0;
  for i:=0 to length(vy)-3 do eq[i]:=0;
  for i:=0 to length(vy)-3 do
  begin
	j:=i+1;
	mat[i,i]:=2/3;
	if i>0 then mat[i,i-1]:=1/6;
	if i<length(vy)-3 then mat[i,i+1]:=1/6;
	eq[i]:=vy[j+1]-2*vy[j]+vy[j-1];
  end;
  wur:=solve(mat,eq);
  splineset.vy:=initv(length(vy));
  for j:=0 to length(vy)-1 do splineset.vy[j]:=vy[j];
  splineset.wur:=initv(length(vy)-2); 
  for j:=0 to length(vy)-3 do splineset.wur[j]:=wur[j];
  prepspline:=splineset;
end;

function getsplineprep(splineset:tsplineset;x:double):double;
function gety(t:double;i:longint):double;
var a,b,c,d,y:double;
begin
  i:=maxi(0,mini(length(splineset.vy)-2,i));
  if (t<i) or (t>(i+1)) then splinewarning:=1;
  b:=t-i;
  a:=1-b;
  y:=a*splineset.vy[i]+b*splineset.vy[i+1];
  if i>0 then 
  begin
    c:=a*(sqr(a)-1)/6;
    y:=y+c*splineset.wur[i-1];
  end;
  if i<length(splineset.vy)-2 then 
  begin
    d:=b*(sqr(b)-1)/6;
    y:=y+d*splineset.wur[i];
  end;
  gety:=y;
end;
begin
  splinewarning:=0;
  getsplineprep:=gety(x,trunc(x));
  if (x<1) or (x>length(splineset.vy)-2) then splinewarning:=1;
  if (x<0) or (x>length(splineset.vy)-1) then splinewarning:=2;
end;

function getspline(vy:tvector;x:double):double;
var splineset:tsplineset;
begin
  splineset:=prepspline(vy);
  getspline:=getsplineprep(splineset,x);
end;

function prepsplinex(vx,vy:tvector):tsplinesetx;
var mat:tmatrix;
	eq,wur:tvector;
    i,j:longint;
	splineset:tsplinesetx;
begin
  if length(vy)<>length(vx) then writeln('getsplinex: x and y vectors have different lengths')
  else
  begin
	mat:=initm(length(vy)-2,length(vy)-2);
	eq:=initv(length(vy)-2);
	wur:=initv(length(vy)-2);

	for j:=0 to length(vy)-3 do for i:=0 to length(vy)-3 do mat[j,i]:=0;
	for i:=0 to length(vy)-3 do eq[i]:=0;
	for i:=0 to length(vy)-3 do
	begin
	  j:=i+1;
	  mat[i,i]:=(vx[j+1]-vx[j-1])/3;
	  if i>0 then mat[i,i-1]:=(vx[j]-vx[j-1])/6;
	  if i<length(vy)-3 then mat[i,i+1]:=(vx[j+1]-vx[j])/6;
	  eq[i]:=(vy[j+1]-vy[j])/(vx[j+1]-vx[j])-(vy[j]-vy[j-1])/(vx[j]-vx[j-1]);
	end;
	wur:=solve(mat,eq);
	splineset.vx:=initv(length(vx));
	for j:=0 to length(vx)-1 do splineset.vx[j]:=vx[j];
	splineset.vy:=initv(length(vy));
	for j:=0 to length(vy)-1 do splineset.vy[j]:=vy[j];
	splineset.wur:=initv(length(vy)-2);
	for j:=0 to length(vy)-3 do splineset.wur[j]:=wur[j];
	prepsplinex:=splineset;
  end;
end;

function getsplineprepx(splineset:tsplinesetx;x:double):double;
var i:longint;
function gety(t:double;i:longint):double;
var a,b,c,d,y:double;
begin
  i:=maxi(0,mini(length(splineset.vy)-2,i));
  if (t<splineset.vx[i]) or (t>splineset.vx[i+1]) then splinewarning:=1;
  b:=(t-splineset.vx[i])/(splineset.vx[i+1]-splineset.vx[i]);
  a:=1-b;
  y:=a*splineset.vy[i]+b*splineset.vy[i+1];
  if i>0 then 
  begin
	c:=a*(sqr(a)-1)/6;
	y:=y+sqr(splineset.vx[i+1]-splineset.vx[i])*c*splineset.wur[i-1];
  end;
  if i<length(splineset.vy)-2 then 
  begin
	d:=b*(sqr(b)-1)/6;
	y:=y+sqr(splineset.vx[i+1]-splineset.vx[i])*d*splineset.wur[i];
  end;
  gety:=y;
end;
begin
  splinewarning:=0;
  i:=0;
  while (i<length(splineset.vx)-1) and (x>splineset.vx[i+1]) do inc(i);
  getsplineprepx:=gety(x,i);
  if (x<splineset.vx[1]) or (x>splineset.vx[length(splineset.vy)-2]) then splinewarning:=1;
  if (x<splineset.vx[0]) or (x>splineset.vx[length(splineset.vy)-1]) then splinewarning:=2;
end;

function getsplinex(vx,vy:tvector;x:double):double;
var splineset:tsplinesetx;
begin
  splineset:=prepsplinex(vx,vy);
  getsplinex:=getsplineprepx(splineset,x);
end;

function getspline2(my:tmatrix;x,y:double):double;
var i,j:longint;
    ux,uy:tvector;
begin
  ux:=initv(length(my[0]));
  uy:=initv(length(my));
  for j:=0 to length(my)-1 do
  begin
    for i:=0 to length(my[0])-1 do ux[i]:=my[j,i];
    uy[j]:=getspline(ux,x);
  end;
  getspline2:=getspline(uy,y);
end;

begin
end.

